import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class RelaisSimulator extends PApplet {


int maxNumber = 5;
int lampSize = 25;
int mode;
boolean pmousePressed;
LampA[] lampsA;
LampB[] lampsB;
LampU[] lampsU;
LampE[] lampsE;
LampPlus switchPlus;
LampMinus switchMinus;
LampPlus asPlus;
LampMinus asMinus;

public void setup(){
  //size(4 * lampSize * (maxNumber + 2), lampSize * 16);
  
  lampsA = new LampA[maxNumber];
  lampsB = new LampB[maxNumber];
  lampsU = new LampU[maxNumber];
  lampsE = new LampE[maxNumber + 1];
  switchPlus = new LampPlus(lampSize, lampSize * 2, lampSize * 5);
  switchPlus.setStatus(true);
  switchMinus = new LampMinus(lampSize, lampSize * 2, lampSize * 7);
  asPlus = new LampPlus(lampSize, lampSize * 2, lampSize * 12);
  asPlus.setUnclickable();
  asMinus = new LampMinus(lampSize, lampSize * 2, lampSize * 14);
  asMinus.setUnclickable();
  for(int i = 0; i < lampsA.length; i ++){
    lampsA[i] = new LampA(lampSize, i * lampSize * 4 + lampSize * 8, lampSize * 3, i);
    lampsA[i].setValue(maxNumber - 1 - i, true);
  }
  for(int i = 0; i < lampsB.length; i ++){
    lampsB[i] = new LampB(lampSize, i * lampSize * 4 + lampSize * 8, lampSize * 6, i);
    lampsB[i].setValue(maxNumber - 1 - i, true);
  }
  for(int i = 0; i < lampsU.length; i ++){
    lampsU[i] = new LampU(lampSize, i * lampSize * 4 + lampSize * 4, lampSize * 9, i);
    lampsU[i].setValue(false);
    lampsU[i].setUnclickable();
  }
  for(int i = 0; i < lampsE.length; i ++){
    lampsE[i] = new LampE(lampSize, i * lampSize * 4 + lampSize * 4, lampSize * 13, i);
    lampsE[i].setValue(maxNumber - i, true);
    lampsE[i].setUnclickable();
  }
  noStroke();
  textAlign(CENTER);
  frameRate(30);
}


public void draw(){
  calculate();
  drawLamps();
  pmousePressed = mousePressed;
}
abstract class AbstractLamp{
  
  int size;
  int xpos, ypos;
  int value;
  int index;
  int lampColor;
  boolean status;
  boolean showDescription;
  boolean clickable;
  String description;
  
  AbstractLamp(int s, int x, int y, int r, int g, int b){
    size = s;
    xpos = x;
    ypos = y;
    lampColor = color(r, g, b);
    clickable = true;
    showDescription = true;
    description = "";
  }
  
  
  public void update(){
    if(mousePressed && !pmousePressed && clickable && pointInCircle(mouseX, mouseY, xpos, ypos, size + 5)){
      status = !status;
    }
  }
  
  
  public void draw(){
    if(status){
      fill(lampColor);
    }else{
      fill(lampColor, 32);
    }
    ellipse(xpos, ypos, size, size);
    if(showDescription){
      fill(200);
      textSize(size / 2);
      text(description, xpos, ypos - size);
    }
  }
  
  
  public void setValue(int v, boolean s){
    value = v;
    description = "2^" + v;
    showDescription = s;
  }
  
  
  public void setValue(boolean s){
    showDescription = s;
  }
  
  
  public void setStatus(boolean s){
    status = s;
  }
  
  
  public void toggleStatus(){
    status = !status;
  }
  
  
  public void setUnclickable(){
    clickable = false;
  }
  
  
  public boolean getStatus(){
    return status;
  }
}
class LampA extends AbstractLamp{
  LampA(int s, int x, int y, int i){
    super(s, x, y, 255, 225, 150);
    index = i;
  }
}
class LampB extends AbstractLamp{
  LampB(int s, int x, int y, int i){
    super(s, x, y, 255, 200, 0);
    index = i;
  }
}
class LampE extends AbstractLamp{
  LampE(int s, int x, int y, int i){
    super(s, x, y, 0, 255, 0);
    index = i;
  }
}
class LampMinus extends AbstractLamp{
  LampMinus(int s, int x, int y){
    super(s, x, y, 0, 255, 0);
    description = "-";
  }
}
class LampPlus extends AbstractLamp{
  LampPlus(int s, int x, int y){
    super(s, x, y, 255, 0, 0);
    description = "+";
  }
}
class LampU extends AbstractLamp{
  LampU(int s, int x, int y, int i){
    super(s, x, y, 255, 0, 0);
  }
}

boolean pSwitchPlus, pSwitchMinus;

public void calculate(){
  updateLamps();
  
  if(mode == 0){
    boolean[][] rowStates = new boolean[2][lampsA.length];
    for(int i = 0; i < rowStates[0].length; i ++){
      rowStates[0][i] = lampsA[i].getStatus();
      rowStates[1][i] = lampsB[i].getStatus();
    }
    boolean[][] result = addition(rowStates);
    for(int i = 0; i < result[0].length - 1; i ++){
      lampsU[i].setStatus(result[0][i]);
    }
    for(int i = 0; i < result[0].length; i ++){
      lampsE[i].setStatus(result[1][i]);
    }
    asPlus.setStatus(true);
    asMinus.setStatus(false);
  }else{
    for(int i = 0; i < lampsU.length; i ++){
      lampsU[i].setStatus(false);
    }
    
    boolean[][] rowStates = new boolean[2][lampsA.length];
    for(int i = 0; i < rowStates[0].length; i ++){
      rowStates[0][i] = lampsA[i].getStatus();
      rowStates[1][i] = lampsB[i].getStatus();
    }
    int firstLampIndex = -1;
    for(int i = 0; i < rowStates[0].length; i ++){
      if(lampsA[rowStates[0].length - i - 1].getStatus() ^ lampsB[rowStates[0].length - i - 1].getStatus()){
        firstLampIndex = rowStates[0].length - i - 1;
      }
    }
    if(firstLampIndex > -1){
      if(lampsA[firstLampIndex].getStatus()){
        asPlus.setStatus(true);
        asMinus.setStatus(false);
        
        for(int i = 0; i < rowStates[0].length; i ++){
          rowStates[1][i] = !rowStates[1][i];
        }
        boolean[][] result = addition(rowStates);
        boolean[][] states = new boolean[2][result[0].length - 1];
        states[1][states[0].length - 1] = result[1][0];
        for(int i = 0; i < states[0].length; i ++){
          states[0][i] = result[1][i + 1];
        }
        
        result = addition(states);
        for(int i = 0; i < lampsE.length; i ++){
          lampsE[i].setStatus(result[1][i]);
        }
      }else{
        asPlus.setStatus(false);
        asMinus.setStatus(true);
        
        for(int i = 0; i < rowStates[0].length; i ++){
          rowStates[1][i] = !rowStates[1][i];
        }
        boolean[][] result = addition(rowStates);
        for(int i = 0; i < result[0].length; i ++){
          result[1][i] = !result[1][i];
        }
        result[1][0] = false;
        for(int i = 0; i < lampsE.length; i ++){
          lampsE[i].setStatus(result[1][i]);
        }
      }
    }else{
      for(int i = 0; i < lampsE.length; i ++){
        lampsE[i].setStatus(false);
      }
      asPlus.setStatus(false);
      asMinus.setStatus(true);
    }
  }
}


public void updateLamps(){
  for(int i = 0; i < lampsA.length; i ++){
    lampsA[i].update();
  }
  for(int i = 0; i < lampsB.length; i ++){
    lampsB[i].update();
  }
  for(int i = 0; i < lampsU.length; i ++){
    lampsU[i].update();
  }
  for(int i = 0; i < lampsE.length; i ++){
    lampsE[i].update();
  }
  
  switchPlus.update();
  switchMinus.update();
  if(switchPlus.getStatus() != pSwitchPlus){
    switchMinus.setStatus(!switchPlus.getStatus());
  }
  if(switchMinus.getStatus() != pSwitchMinus){
    switchPlus.setStatus(!switchMinus.getStatus());
  }
  if(switchPlus.getStatus()){
    mode = 0;
  }else{
    mode = 1;
  }
  asPlus.update();
  asMinus.update();
  pSwitchPlus = switchPlus.getStatus();
  pSwitchMinus = switchMinus.getStatus();
}


public boolean[][] addition(boolean[][] input){
  boolean[][] output = new boolean[2][input[0].length + 1];
  int maxI = output[0].length - 1;
  for(int i = 0; i < maxI + 1; i ++){
    int numberOfActiveBits;
    if(i < maxI){
      numberOfActiveBits = PApplet.parseInt(input[0][maxI - i - 1]) + PApplet.parseInt(input[1][maxI - i - 1]) + PApplet.parseInt(output[0][maxI - i]);
      if(numberOfActiveBits % 2 == 1){
        output[1][maxI - i] = true;
      }else{
        output[1][maxI - i] = false;
      }
      if(numberOfActiveBits > 1){
        output[0][maxI - i - 1] = true;
      }else{
        output[0][maxI - i - 1] = false;
      }
    }else{
      output[1][0] = output[0][0];
    }
  }
  return output;
}
public void drawLamps(){
  background(32);
  for(int i = 0; i < lampsA.length; i ++){
    lampsA[i].draw();
  }
  for(int i = 0; i < lampsB.length; i ++){
    lampsB[i].draw();
  }
  for(int i = 0; i < lampsU.length; i ++){
    lampsU[i].draw();
  }
  for(int i = 0; i < lampsE.length; i ++){
    lampsE[i].draw();
  }
  switchPlus.draw();
  switchMinus.draw();
  asPlus.draw();
  asMinus.draw();
}
public boolean pointInCircle(int pointX, int pointY, int centerX, int centerY, int size){
  if(sqrt(sq(pointX - centerX) + sq(pointY - centerY)) < size / 2){
    return true;
  }
  return false;
}
  public void settings() {  size(700, 400); }
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "RelaisSimulator" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
